#include "SceneGraph.h"
#include "Exporter.h"
#include "UnigraphicsParser.h"
#include "X3dParser.h"
#include "Preprocessor.h"
#include "ViewCell.h"
#include "Environment.h"
#include "ViewCellsManager.h"
#include "ViewCellBsp.h"
#include "VspBspTree.h"
#include "RenderSimulator.h"
#include "GlRenderer.h"
#include "PlyParser.h"
#include "SamplingStrategy.h"
#include "VspTree.h"
#include "OspTree.h"
#include "ObjParser.h"
#include "BvHierarchy.h"

#ifdef GTP_INTERNAL
#include "ArchModeler2MLRT.hxx"
#endif

namespace GtpVisibilityPreprocessor {

const static bool ADDITIONAL_GEOMETRY_HACK = false;

//Preprocessor *preprocessor;


// HACK: Artificially modify scene to watch rendercost changes
static void AddGeometry(SceneGraph *scene)
{
	scene->mRoot->UpdateBox();

	AxisAlignedBox3 sceneBox = scene->GetBox();

	int n = 200;

	if (0){
	// form grid of boxes
	for (int i = 0; i < n; ++ i)
	{
		for (int j = 0; j < n; ++ j)
		{
			const Vector3 scale2((float)j * 0.8f / n + 0.1f,  0.05f, (float)i * 0.8f  / (float)n + 0.1f);
		
			const Vector3 pt2 = sceneBox.Min() + scale2 * (sceneBox.Max() - sceneBox.Min());
		
			const Vector3 boxSize = sceneBox.Size() * Vector3(0.0025f, 0.01f, 0.0025f);
			AxisAlignedBox3 box(pt2, pt2 + boxSize);
			Mesh *mesh = CreateMeshFromBox(box);

			mesh->Preprocess();
		
			MeshInstance *mi = new MeshInstance(mesh);
			scene->mRoot->mGeometry.push_back(mi);
		}
	}

	for (int i = 0; i < n; ++ i)
	{
		for (int j = 0; j < n; ++ j)
		{
			const Vector3 scale2(0.15f, (float)j * 0.8f / n + 0.1f, (float)i * 0.8f  / (float)n + 0.1f);
		
			Vector3 pt2 = sceneBox.Min() + scale2 * (sceneBox.Max() - sceneBox.Min());
		
			Vector3 boxSize = sceneBox.Size() * Vector3(0.0025f, 0.01f, 0.0025f);
			AxisAlignedBox3 box(pt2, pt2 + boxSize);
			Mesh *mesh = CreateMeshFromBox(box);

			mesh->Preprocess();
		
			MeshInstance *mi = new MeshInstance(mesh);
			scene->mRoot->mGeometry.push_back(mi);
		}
	}

	for (int i = 0; i < n; ++ i)
	{
		const Vector3 scale2(2, 0.2f, (float)i * 0.8f  / (float)n + 0.1f);
		
		Vector3 pt2 = sceneBox.Min() + scale2 * (sceneBox.Max() - sceneBox.Min());
		
		//Vector3 boxSize = sceneBox.Size() * Vector3(0.0025, 0.01, 0.0025);
		Vector3 boxSize = sceneBox.Size() * Vector3(0.005f, 0.02f, 0.005f);

		AxisAlignedBox3 box(pt2 + 0.1f, pt2 + boxSize);
		Mesh *mesh = CreateMeshFromBox(box);

		mesh->Preprocess();
		
		MeshInstance *mi = new MeshInstance(mesh);
		scene->mRoot->mGeometry.push_back(mi);
	}
	
	scene->mRoot->UpdateBox();
	}
		
	if (1)
	{
		// plane separating view space regions
		const Vector3 scale(1.0f, 0.0, 0);

		Vector3 pt = sceneBox.Min() + scale * (sceneBox.Max() - sceneBox.Min());

		Plane3 cuttingPlane(Vector3(1, 0, 0), pt);
		Mesh *planeMesh = new Mesh();
		
		Polygon3 *poly = sceneBox.CrossSection(cuttingPlane);
		IncludePolyInMesh(*poly, *planeMesh);
		
		planeMesh->Preprocess();
		
		MeshInstance *planeMi = new MeshInstance(planeMesh);
		scene->mRoot->mGeometry.push_back(planeMi);
	}	
}


Preprocessor::Preprocessor():
mKdTree(NULL),
mBspTree(NULL),
mVspBspTree(NULL),
mVspTree(NULL),
mOspTree(NULL),
mBvHierarchy(NULL),
mViewCellsManager(NULL),
mRenderSimulator(NULL),
mPass(0)
{
	Environment::GetSingleton()->GetBoolValue("Preprocessor.useGlRenderer", mUseGlRenderer);
  
	// renderer will be constructed when the scene graph and viewcell manager will be known
	renderer = NULL;
  
	Environment::GetSingleton()->GetBoolValue("Preprocessor.useGlDebugger", mUseGlDebugger);
	Environment::GetSingleton()->GetBoolValue("Preprocessor.loadPolygonsAsMeshes", mLoadPolygonsAsMeshes);
	Environment::GetSingleton()->GetBoolValue("Preprocessor.quitOnFinish", mQuitOnFinish);
	Environment::GetSingleton()->GetBoolValue("Preprocessor.computeVisibility", mComputeVisibility);
	Environment::GetSingleton()->GetBoolValue("Preprocessor.detectEmptyViewSpace", mDetectEmptyViewSpace);
	Environment::GetSingleton()->GetBoolValue("Preprocessor.exportVisibility", mExportVisibility );
	Environment::GetSingleton()->GetIntValue("Preprocessor.rayCastMethod", mRayCastMethod);

	char buffer[256];
	Environment::GetSingleton()->GetStringValue("Preprocessor.visibilityFile",  buffer);
	mVisibilityFileName = buffer;
	Environment::GetSingleton()->GetBoolValue("Preprocessor.applyVisibilityFilter", mApplyVisibilityFilter );
	Environment::GetSingleton()->GetBoolValue("Preprocessor.applyVisibilitySpatialFilter",
							  mApplyVisibilitySpatialFilter );
	Environment::GetSingleton()->GetFloatValue("Preprocessor.visibilityFilterWidth", mVisibilityFilterWidth);

	Debug << "detect empty view space=" << mDetectEmptyViewSpace << endl;
	Debug << "load polygons as meshes: " << mLoadPolygonsAsMeshes << endl;
}


Preprocessor::~Preprocessor()
{
  cout << "cleaning up" << endl;

  cout << "Deleting view cells manager ... \n";
  DEL_PTR(mViewCellsManager);
  cout << "done.\n";

  cout << "Deleting bsp tree ... \n";
  DEL_PTR(mBspTree);
  cout << "done.\n";

  cout << "Deleting kd tree...\n";
  DEL_PTR(mKdTree);
  cout << "done.\n";
 
  cout << "Deleting vsp tree...\n";
  DEL_PTR(mVspTree);
  cout << "done.\n";

  cout << "Deleting osp tree...\n";
  DEL_PTR(mOspTree);
  cout << "done.\n";

  cout << "Deleting vspbsp tree...\n";
  DEL_PTR(mVspBspTree);
  cout << "done.\n";

   cout << "Deleting scene graph...\n";
  DEL_PTR(mSceneGraph);
  cout << "done.\n";

  DEL_PTR(mRenderSimulator);
  DEL_PTR(renderer);
}

int
SplitFilenames(const string str, vector<string> &filenames)
{
	int pos = 0;

	while(1) {
		int npos = (int)str.find(';', pos);
		
		if (npos < 0 || npos - pos < 1)
			break;
		filenames.push_back(string(str, pos, npos - pos));
		pos = npos + 1;
	}
	
	filenames.push_back(string(str, pos, str.size() - pos));
	return (int)filenames.size();
}


bool
Preprocessor::LoadScene(const string filename)
{
	// use leaf nodes of the original spatial hierarchy as occludees
	mSceneGraph = new SceneGraph;
  
	Parser *parser;
	vector<string> filenames;
	int files = SplitFilenames(filename, filenames);
	cout << "number of input files: " << files << endl;
	bool result = false;
	if (files == 1) {
		
		if (strstr(filename.c_str(), ".x3d"))
		  parser = new X3dParser;
		else
		  if (strstr(filename.c_str(), ".ply") || strstr(filename.c_str(), ".plb"))
			parser = new PlyParser;
		  else if (strstr(filename.c_str(), ".obj"))
			  parser = new ObjParser;
		  else 
			  parser = new UnigraphicsParser;

		cout<<filename<<endl;

		if (mRayCastMethod == Preprocessor::INTEL_RAYCASTER)
		  result = parser->ParseFile(filename, &mSceneGraph->mRoot,
									 mLoadPolygonsAsMeshes,
									 &mFaceParents);
		else
		  result = parser->ParseFile(filename, &mSceneGraph->mRoot, mLoadPolygonsAsMeshes);
		
		delete parser;

	} else {
		// root for different files
		mSceneGraph->mRoot = new SceneGraphNode;
		for (int i= 0; i < filenames.size(); i++) {
		  if (strstr(filenames[i].c_str(), ".x3d"))
			parser = new X3dParser;
		  else 
			parser = new UnigraphicsParser;
		  
		  SceneGraphNode *node;
		  bool success;
		  
		  if (mRayCastMethod == Preprocessor::INTEL_RAYCASTER)
			success = parser->ParseFile(filename, &node,
										mLoadPolygonsAsMeshes,
										&mFaceParents);
		  else
			success = parser->ParseFile(filename, &node, mLoadPolygonsAsMeshes);
		  
		  if (success) {
			mSceneGraph->mRoot->mChildren.push_back(node);
			// at least one file parsed
			result = true;
		  }
		  
		  delete parser;
		}
	}
	
	
	if (result) 
	  {
		// HACK 
		if (ADDITIONAL_GEOMETRY_HACK)
			AddGeometry(mSceneGraph);
		
		mSceneGraph->AssignObjectIds();
	
		int intersectables, faces;
		mSceneGraph->GetStatistics(intersectables, faces);
        
		cout<<filename<<" parsed successfully."<<endl;
		cout<<"#NUM_OBJECTS (Total numner of objects)\n"<<intersectables<<endl;
		cout<<"#NUM_FACES (Total numner of faces)\n"<<faces<<endl;
		mSceneGraph->CollectObjects(&mObjects);
		mSceneGraph->mRoot->UpdateBox();

		if (0)
		{
			Exporter *exporter = Exporter::GetExporter("testload.x3d");

			if (exporter)
			{
				exporter->ExportGeometry(mObjects);
				delete exporter;
			}
		}
	}
	
	
	return result;
}

bool
Preprocessor::ExportPreprocessedData(const string filename)
{
  
  mViewCellsManager->ExportViewCells(filename, true, mObjects);
  
  return true;
}

bool
Preprocessor::PostProcessVisibility()
{
  
  if (mApplyVisibilityFilter || mApplyVisibilitySpatialFilter) {
	cout<<"Applying visibility filter ...";
	cout<<"filter width = " << mVisibilityFilterWidth << endl;
	
	if (!mViewCellsManager)
	  return false;
	
	mViewCellsManager->ApplyFilter(mKdTree,
								   mApplyVisibilityFilter ? mVisibilityFilterWidth : -1.0f,
								   mApplyVisibilitySpatialFilter ? mVisibilityFilterWidth : -1.0f);
	cout << "done." << endl;
  }
  
  // export the preprocessed information to a file
  if (mExportVisibility)
	ExportPreprocessedData(mVisibilityFileName);
  
  return true;
}


bool
Preprocessor::BuildKdTree()
{
  mKdTree = new KdTree;
  // add mesh instances of the scene graph to the root of the tree
  KdLeaf *root = (KdLeaf *)mKdTree->GetRoot();
  mSceneGraph->CollectObjects(&root->mObjects);
  
  long startTime = GetTime();

  cout << "building kd tree ... " << endl;
  mKdTree->Construct();
  cout << "construction finished in " << TimeDiff(startTime, GetTime()) * 1e-3 << " secs " << endl;
  return true;
}

void
Preprocessor::KdTreeStatistics(ostream &s)
{
  s<<mKdTree->GetStatistics();
}

void
Preprocessor::BspTreeStatistics(ostream &s)
{
	s << mBspTree->GetStatistics();
}

bool
Preprocessor::Export( const string filename,
					  const bool scene,
					  const bool kdtree,
					  const bool bsptree
					  )
{
  Exporter *exporter = Exporter::GetExporter(filename);
	
  if (exporter) {
    if (scene)
      exporter->ExportScene(mSceneGraph->mRoot);

    if (kdtree) {
      exporter->SetWireframe();
      exporter->ExportKdTree(*mKdTree);
    }

	if (bsptree) {
		//exporter->SetWireframe();
		exporter->ExportBspTree(*mBspTree);
	}

    delete exporter;
    return true;
  }

  return false;
}


bool Preprocessor::PrepareViewCells()
{
	//-- parse view cells construction method
	Environment::GetSingleton()->GetBoolValue("ViewCells.loadFromFile", mLoadViewCells);
	char buf[100];
	
	if (mLoadViewCells)
	{	
		Environment::GetSingleton()->GetStringValue("ViewCells.filename", buf);
		cout << "loading view cells from " << buf << endl;
		mViewCellsManager = ViewCellsManager::LoadViewCells(buf, &mObjects, true);
	}
	else
	{
		//-- parse type of view cell container

		Environment::GetSingleton()->GetStringValue("ViewCells.type", buf);		
	    mViewCellsManager = CreateViewCellsManager(buf);

		// default view space is the extent of the scene
		mViewCellsManager->SetViewSpaceBox(mSceneGraph->GetBox());


	}
	
	//-- parameters for render heuristics evaluation
	float objRenderCost = 0, vcOverhead = 0, moveSpeed = 0;

	Environment::GetSingleton()->GetFloatValue("Simulation.objRenderCost",objRenderCost);
	Environment::GetSingleton()->GetFloatValue("Simulation.vcOverhead", vcOverhead);
	Environment::GetSingleton()->GetFloatValue("Simulation.moveSpeed", moveSpeed);
	
	mRenderSimulator = 
		new RenderSimulator(mViewCellsManager, objRenderCost, vcOverhead, moveSpeed);

	mViewCellsManager->SetRenderer(mRenderSimulator);


	if (mUseGlRenderer || mUseGlDebugger)
	{
		// NOTE: render texture should be power of 2 and square
		// renderer must be initialised
		// $$matt
//		renderer = new GlRendererBuffer(1024, 768, mSceneGraph, mViewCellsManager, mKdTree);
		//		renderer->makeCurrent();
		
	}
	
	return true;
}


ViewCellsManager *Preprocessor::CreateViewCellsManager(const char *name)
{
	ViewCellsTree *vcTree = new ViewCellsTree;

	if (strcmp(name, "kdTree") == 0)
	{
		mViewCellsManager = new KdViewCellsManager(vcTree, mKdTree);
	}
	else if (strcmp(name, "bspTree") == 0)
	{
		Debug << "view cell type: Bsp" << endl;

		mBspTree = new BspTree();
		mViewCellsManager = new BspViewCellsManager(vcTree, mBspTree);
	}
	else if (strcmp(name, "vspBspTree") == 0)
	{
		Debug << "view cell type: VspBsp" << endl;

		mVspBspTree = new VspBspTree();
		mViewCellsManager = new VspBspViewCellsManager(vcTree, mVspBspTree);
	}
	else if (strcmp(name, "vspOspTree") == 0)
	{
		mVspTree = new VspTree();
		//mOspTree = new OspTree();
		
		// HACK for testing if per kd evaluation works!!
		mOspTree = new OspTree(*mKdTree);

		mViewCellsManager = new VspOspViewCellsManager(vcTree, mVspTree, mOspTree);
	}
	else if (strcmp(name, "sceneDependent") == 0)
	{
		Debug << "view cell type: Bsp" << endl;

		//TODO
		mBspTree = new BspTree();
		mViewCellsManager = new BspViewCellsManager(vcTree, mBspTree);
	}
	else
	{
		cerr << "Wrong view cells type " << name << "!!!" << endl;
		exit(1);
	}

	//vcTree->SetViewCellsManager(mViewCellsManager);
	return mViewCellsManager;
}


// use ascii format to store rays
#define USE_ASCII 0


static inline bool ilt(Intersectable *obj1, Intersectable *obj2)
{
	return obj1->mId < obj2->mId;
}


bool Preprocessor::LoadKdTree()
{
	return true;
}

bool Preprocessor::ExportKdTree()
{
	return true;
}


bool Preprocessor::LoadSamples(VssRayContainer &samples, 
							   ObjectContainer &objects) const
{
	std::stable_sort(objects.begin(), objects.end(), ilt);
	char fileName[100];
	Environment::GetSingleton()->GetStringValue("Preprocessor.samplesFilename", fileName);
	
    Vector3 origin, termination;
	// HACK: needed only for lower_bound algorithm to find the 
	// intersected objects
	MeshInstance sObj(NULL);
	MeshInstance tObj(NULL);

#if USE_ASCII
	ifstream samplesIn(fileName);
	if (!samplesIn.is_open())
		return false;

	string buf;
	while (!(getline(samplesIn, buf)).eof())
	{
		sscanf(buf.c_str(), "%f %f %f %f %f %f %d %d", 
			   &origin.x, &origin.y, &origin.z,
			   &termination.x, &termination.y, &termination.z, 
			   &(sObj.mId), &(tObj.mId));
		
		Intersectable *sourceObj = NULL;
		Intersectable *termObj = NULL;
		
		if (sObj.mId >= 0)
		{
			ObjectContainer::iterator oit =
				lower_bound(objects.begin(), objects.end(), &sObj, ilt);
			sourceObj = *oit;
		}
		
		if (tObj.mId >= 0)
		{
			ObjectContainer::iterator oit =
				lower_bound(objects.begin(), objects.end(), &tObj, ilt);
			termObj = *oit;
		}

		samples.push_back(new VssRay(origin, termination, sourceObj, termObj));
	}
#else
	ifstream samplesIn(fileName, ios::binary);
	if (!samplesIn.is_open())
		return false;

	while (1)
	{
		 samplesIn.read(reinterpret_cast<char *>(&origin), sizeof(Vector3));
		 samplesIn.read(reinterpret_cast<char *>(&termination), sizeof(Vector3));
		 samplesIn.read(reinterpret_cast<char *>(&(sObj.mId)), sizeof(int));
		 samplesIn.read(reinterpret_cast<char *>(&(tObj.mId)), sizeof(int));
		
		 if (samplesIn.eof())
			break;

		Intersectable *sourceObj = NULL;
		Intersectable *termObj = NULL;
		
		if (sObj.mId >= 0)
		{
			ObjectContainer::iterator oit =
				lower_bound(objects.begin(), objects.end(), &sObj, ilt);
			sourceObj = *oit;
		}
		
		if (tObj.mId >= 0)
		{
			ObjectContainer::iterator oit =
				lower_bound(objects.begin(), objects.end(), &tObj, ilt);
			termObj = *oit;
		}

		samples.push_back(new VssRay(origin, termination, sourceObj, termObj));
	}

#endif
	samplesIn.close();

	return true;
}


bool Preprocessor::ExportSamples(const VssRayContainer &samples) const
{
	char fileName[100];
	Environment::GetSingleton()->GetStringValue("Preprocessor.samplesFilename", fileName);
	

	VssRayContainer::const_iterator it, it_end = samples.end();
	
#if USE_ASCII
	ofstream samplesOut(fileName);
	if (!samplesOut.is_open())
		return false;

	for (it = samples.begin(); it != it_end; ++ it)
	{
		VssRay *ray = *it;
		int sourceid = ray->mOriginObject ? ray->mOriginObject->mId : -1;		
		int termid = ray->mTerminationObject ? ray->mTerminationObject->mId : -1;	

		samplesOut << ray->GetOrigin().x << " " << ray->GetOrigin().y << " " << ray->GetOrigin().z << " " 
				   << ray->GetTermination().x << " " << ray->GetTermination().y << " " << ray->GetTermination().z << " " 
				   << sourceid << " " << termid << "\n";
	}
#else
	ofstream samplesOut(fileName, ios::binary);
	if (!samplesOut.is_open())
		return false;

	for (it = samples.begin(); it != it_end; ++ it)
	{	
		VssRay *ray = *it;
		Vector3 origin(ray->GetOrigin());
		Vector3 termination(ray->GetTermination());
		
		int sourceid = ray->mOriginObject ? ray->mOriginObject->mId : -1;		
		int termid = ray->mTerminationObject ? ray->mTerminationObject->mId : -1;		

		samplesOut.write(reinterpret_cast<char *>(&origin), sizeof(Vector3));
		samplesOut.write(reinterpret_cast<char *>(&termination), sizeof(Vector3));
		samplesOut.write(reinterpret_cast<char *>(&sourceid), sizeof(int));
		samplesOut.write(reinterpret_cast<char *>(&termid), sizeof(int));
    }
#endif
	samplesOut.close();

	return true;
}

#if 0 // matt: implemented interface samplestrategy
bool
Preprocessor::GenerateRays(
						   const int number,
						   const int sampleType,
						   SimpleRayContainer &rays
						   )
{
  Vector3 origin, direction; 
  int startSize = (int)rays.size();
  for (int i=0; (int)rays.size() - startSize  < number; i ++) {
	// now get the direction
	switch (sampleType) {
	case OBJECT_BASED_DISTRIBUTION: {
	  mViewCellsManager->GetViewPoint(origin);
	  Vector3 point;
	  Vector3 normal;
	  int i = RandomValue(0, mObjects.size() - 1);
	  Intersectable *object = mObjects[i];
	  object->GetRandomSurfacePoint(point, normal);
	  direction = point - origin;
	}
	  break;
	case OBJECT_DIRECTION_BASED_DISTRIBUTION: {
	  int i = RandomValue(0, mObjects.size() - 1);
	  Intersectable *object = mObjects[i];
	  Vector3 normal;
	  object->GetRandomSurfacePoint(origin, normal);
	  direction = UniformRandomVector(normal);
	  origin += 0.1f*direction;
	}
	  break;
	case DIRECTION_BASED_DISTRIBUTION:
	  mViewCellsManager->GetViewPoint(origin);
	  direction = UniformRandomVector();
	  break;
	case DIRECTION_BOX_BASED_DISTRIBUTION: {
	  mViewCellsManager->GetViewPoint(origin);
	  float alpha = RandomValue(0.0f, 2*M_PI);
	  float beta = RandomValue(-M_PI/2, M_PI/2);
	  direction = VssRay::GetDirection(alpha, beta);
	  break;
	}
	case SPATIAL_BOX_BASED_DISTRIBUTION:
	  mViewCellsManager->GetViewPoint(origin);
	  direction = mKdTree->GetBox().GetRandomPoint() - origin;
	  break;
	default:
	  // unsuported distribution type
	  return false;
	}
	// $$ jb the pdf is yet not correct for all sampling methods!
	float pdf = 1.0f;
	float c = Magnitude(direction);
	if (c > Limits::Small) {
	  direction*=1.0f/c;
	  rays.AddRay(SimpleRay(origin, direction, pdf));
	}
  }
  return true;
}
#endif
bool Preprocessor::GenerateRays(const int number,
								const int sampleType,
								SimpleRayContainer &rays)
{
	Vector3 origin, direction; 
	
	const int startSize = (int)rays.size();
	SamplingStrategy *strategy = GenerateSamplingStrategy(sampleType);

	if (!strategy)
		return false;

	for (int i=0; (int)rays.size() - startSize < number; ++ i) 
	{
		SimpleRay newRay;
		bool success = strategy->GenerateSample(newRay);

		if (success)
			rays.AddRay(newRay);
	}

	delete strategy;

    return true;
}


SamplingStrategy *Preprocessor::GenerateSamplingStrategy(const int strategyId) const
{
	switch (strategyId)
	{
	case OBJECT_BASED_DISTRIBUTION: 
		return new ObjectBasedDistribution(*this);
	case OBJECT_DIRECTION_BASED_DISTRIBUTION:
		return new ObjectDirectionBasedDistribution(*this);
	case DIRECTION_BASED_DISTRIBUTION:
		return new DirectionBasedDistribution(*this);
	case DIRECTION_BOX_BASED_DISTRIBUTION: 
		return new DirectionBoxBasedDistribution(*this);
	case SPATIAL_BOX_BASED_DISTRIBUTION:
		return new SpatialBoxBasedDistribution(*this);
	//case OBJECTS_INTERIOR_DISTRIBUTION:
	//	return new ObjectsInteriorDistribution(*this);
	default: // no valid strategy
		return NULL;
	}

	// should never come here
	return NULL;
}


bool Preprocessor::InitRayCast(const string externKdTree)
{
	switch (mRayCastMethod) // use intel ray tracing
	{
	case INTEL_RAYCASTER:
#ifdef GTP_INTERNAL
	  cout<<"Ray Cast file: "<<externKdTree<<endl;
	  return mlrtaLoadAS(externKdTree.c_str());
#endif
	case INTERNAL_RAYCASTER:
	default:
		break;
	}

	return true;
}


int Preprocessor::CastIntelDoubleRay(
									 const Vector3 &viewPoint,
									 const Vector3 &direction,
									 const float probability,
									 VssRayContainer &vssRays,
									 const AxisAlignedBox3 &box
									 )
{
	VssRay *vssRay  = NULL;
	int hits = 0;

	Vector3 pointA, pointB;
	
	Intersectable *objectA = 
		CastIntelSingleRay(viewPoint, direction, pointA, box);
	
	// cast ray into other direction
	Intersectable *objectB = 
		CastIntelSingleRay(viewPoint, -direction, pointB, box);

	const bool validSample = (objectA != objectB);
	
	if (validSample) 
	{	
		if (objectA) 
		{
			vssRay = new VssRay(pointB,
								pointA,
								objectB,
								objectA,
								probability,
								mPass);

			vssRays.push_back(vssRay);
			hits ++;
		}

		if (objectB) 
		{
			vssRay = new VssRay(pointA,
								pointB,
								objectA,
								objectB,
								probability,
								mPass);

			vssRays.push_back(vssRay);
			hits ++;
		}
		//Debug << "intel ray: " << *vssRay << endl;
	}
	//cout << "a";
	return hits;
}


Intersectable *Preprocessor::CastIntelSingleRay(const Vector3 &viewPoint,
												const Vector3 &direction,
												//const float probability,
												Vector3 &tPoint,
												const AxisAlignedBox3 &box
												)
{
	AxisAlignedBox3 sbox = box;
	sbox.Enlarge(Vector3(-Limits::Small));

	if (!sbox.IsInside(viewPoint))
		return 0;
	
	float pforg[3];
	float pfdir[3];
	double pfnorm[3];

	pforg[0] = viewPoint[0]; pforg[1] = viewPoint[1]; pforg[2] = viewPoint[2];
	pfdir[0] = direction[0]; pfdir[1] = direction[1]; pfdir[2] = direction[2];

	float dist = 0;
#ifdef GTP_INTERNAL
	const int hittriangle = mlrtaIntersectAS(pforg, pfdir, pfnorm, dist);
#else
	const int hittriangle = -1;
#endif

	if (hittriangle == -1)
	{
		static Ray ray;
		SetupRay(ray, viewPoint, direction);

		float tmin = 0, tmax;
		if (box.ComputeMinMaxT(ray, &tmin, &tmax) && tmin < tmax)
		{
			tPoint = ray.Extrap(tmax);
		}

		return NULL;
	}
	else
	{
		tPoint[0] = pforg[0] + pfdir[0] * dist;
		tPoint[1] = pforg[1] + pfdir[1] * dist;
		tPoint[2] = pforg[2] + pfdir[2] * dist;

		return mFaceParents[hittriangle];
	}
}


int Preprocessor::CastInternalRay(
								  const Vector3 &viewPoint,
								  const Vector3 &direction,
								  const float probability,
								  VssRayContainer &vssRays,
								  const AxisAlignedBox3 &box
								  )
{

  int hits = 0;
  static Ray ray;
  Intersectable *objectA, *objectB;
  Vector3 pointA, pointB;

  //  AxisAlignedBox3 box = Union(mKdTree->GetBox(), mViewCellsManager->GetViewSpaceBox());
  

  AxisAlignedBox3 sbox = box;
  sbox.Enlarge(Vector3(-Limits::Small));
  if (!sbox.IsInside(viewPoint))
	return 0;
	
  SetupRay(ray, viewPoint, direction);
  ray.mFlags &= ~Ray::CULL_BACKFACES;

  // cast ray to KD tree to find intersection with other objects
  float bsize = Magnitude(box.Size());
  
  
  if (mKdTree->CastRay(ray)) {
	objectA = ray.intersections[0].mObject;
	pointA = ray.Extrap(ray.intersections[0].mT);
	if (mDetectEmptyViewSpace)
	  if (DotProd(ray.intersections[0].mNormal, direction) >= 0) {
		// discard the sample
		return 0;
	  }
	
  } else {
	objectA = NULL;
	// compute intersection with the scene bounding box
	float tmin, tmax;
	if (box.ComputeMinMaxT(ray, &tmin, &tmax) && tmin < tmax)
	  pointA = ray.Extrap(tmax);
	else
	  return 0;
  }

  
  SetupRay(ray, viewPoint, -direction);
  ray.mFlags &= ~Ray::CULL_BACKFACES;
  
  if (mKdTree->CastRay(ray)) {
	objectB = ray.intersections[0].mObject;
	pointB = ray.Extrap(ray.intersections[0].mT);
	if (mDetectEmptyViewSpace)
	  if (DotProd(ray.intersections[0].mNormal, direction) <= 0) {
		// discard the sample
		return 0;
	  }
  } else {
	objectB = NULL;
	float tmin, tmax;
	if (box.ComputeMinMaxT(ray, &tmin, &tmax) && tmin < tmax)
	  pointB = ray.Extrap(tmax);
	else
	  return 0;
  }
  
  
  VssRay *vssRay  = NULL;
  bool validSample = (objectA != objectB);
  if (validSample) {
	if (objectA) {
	  vssRay = new VssRay(pointB,
						  pointA,
						  objectB,
						  objectA,
						  mPass,
						  probability
						  );
	  vssRays.push_back(vssRay);
	  hits ++;
	}
	
	if (objectB) {
	  vssRay = new VssRay(pointA,
						  pointB,
						  objectA,
						  objectB,
						  mPass,
						  probability
						  );
	  vssRays.push_back(vssRay);
	  hits ++;
	}
  }
  
  return hits;
}


int Preprocessor::CastRay(
						  const Vector3 &viewPoint,
						  const Vector3 &direction,
						  const float probability,
						  VssRayContainer &vssRays,
						  const AxisAlignedBox3 &box
						  )
{
	switch (mRayCastMethod)
	{
	case INTEL_RAYCASTER:
		return CastIntelDoubleRay(viewPoint, direction, probability, vssRays, box);
	case INTERNAL_RAYCASTER:
	default:
		return CastInternalRay(viewPoint, direction, probability, vssRays, box);
	}
}


void Preprocessor::SetupRay(Ray &ray,
							const Vector3 &point,
							const Vector3 &direction
							)
{
	ray.Clear();
	// do not store anything else then intersections at the ray
	ray.Init(point, direction, Ray::LOCAL_RAY);
}

}
